"""Schemas for smart logger internals."""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Any, List
from uuid import UUID

from pybotx.models.commands import SystemEvent


class LogSourceType(str, Enum):
    USER_MESSAGE = "USER_MESSAGE"
    ASYNCIO_TASK = "ASYNCIO_TASK"
    HTTP_HANDLER = "HTTP_HANDLER"
    SYSTEM_EVENT = "SYSTEM_EVENT"


@dataclass
class LogSource:
    source_type: LogSourceType
    source_context: Any

    @classmethod
    def from_user_message(cls, user_huid: UUID) -> LogSource:
        return cls(source_type=LogSourceType.USER_MESSAGE, source_context=user_huid)

    @classmethod
    def from_asyncio_task(cls, task_name: str) -> LogSource:
        return cls(source_type=LogSourceType.ASYNCIO_TASK, source_context=task_name)

    @classmethod
    def from_http_request(cls, method: str, url: str) -> LogSource:
        return cls(source_type=LogSourceType.HTTP_HANDLER, source_context=(method, url))

    @classmethod
    def from_system_event(cls, event: SystemEvent) -> LogSource:
        return cls(
            source_type=LogSourceType.SYSTEM_EVENT,
            source_context=(event.bot.id, event.__class__.__name__),
        )


@dataclass
class LogEntry:
    log_message: str
    log_item: Any
    module: str
    function: str
    line_number: int


AccumulatedLogs = List[LogEntry]



"""Helpers for printing log messages."""

from pprint import pformat
from typing import Any, Dict

from pybotx.logger import trim_file_data_in_incoming_json
from loguru import logger

from pybotx_smart_logger.contextvars import get_http_request_headers, get_log_source
from pybotx_smart_logger.schemas import LogSourceType

MAX_FILE_CONTENT_LENGTH = 32


def attach_log_source(log_message: str) -> str:
    raw_log_source = get_log_source()

    if raw_log_source is None:
        log_source = "<missing log source>"
    elif raw_log_source.source_type == LogSourceType.USER_MESSAGE.value:
        log_source = f"user '{raw_log_source.source_context}'"
    elif raw_log_source.source_type == LogSourceType.ASYNCIO_TASK.value:
        log_source = f"task '{raw_log_source.source_context}'"
    elif raw_log_source.source_type == LogSourceType.HTTP_HANDLER.value:
        method, url = raw_log_source.source_context
        log_source = f"HTTP {method} {url}"
    elif raw_log_source.source_type == LogSourceType.SYSTEM_EVENT.value:
        bot_id, event = raw_log_source.source_context
        log_source = f"event {event} to bot '{bot_id}'"
    else:
        log_source = "<UNKNOWN LOG SOURCE>"

    return f"[{log_source}] {log_message}"


def log_incoming_message(
    raw_command: Dict[str, Any], title: str, log_level: str
) -> None:
    trimmed_raw_command = trim_file_data_in_incoming_json(raw_command)
    logger.log(log_level, "{}\n{}", title, pformat(trimmed_raw_command))


def log_system_event(raw_command: Dict[str, Any], title: str, log_level: str) -> None:
    logger.log(log_level, "{}\n{}", title, pformat(raw_command))


def log_incoming_http_request(
    title: str,
    log_level: str,
) -> None:
    log_source = get_log_source()
    assert log_source is not None
    method, url = log_source.source_context

    headers = get_http_request_headers()
    assert headers is not None

    formatted_headers = "\n".join([f"{key}: {value}" for key, value in headers.items()])
    logger.log(
        log_level,
        "HTTP {} {}\nHeaders:\n{}",
        method,
        url,
        formatted_headers,
    )



"""Exception handler for smart_logger.

Prints all accumulated logs and sends default error message.
"""

from typing import Awaitable, Callable

from pybotx import Bot, IncomingMessage
from pybotx_smartapp_rpc.models.errors import RPCError
from pybotx_smartapp_rpc.models.responses import RPCErrorResponse
from pybotx_smartapp_rpc.smartapp import SmartApp
from fastapi import Request
from loguru import logger
from starlette.responses import Response
from starlette.status import HTTP_500_INTERNAL_SERVER_ERROR

from pybotx_smart_logger import log_levels
from pybotx_smart_logger.contextvars import get_debug_enabled
from pybotx_smart_logger.logger import flush_accumulated_logs
from pybotx_smart_logger.output import (
    attach_log_source,
    log_system_event,
    log_incoming_http_request,
    log_incoming_message,
)


def make_smart_logger_exception_handler(
    error_text: str,
) -> Callable[[Exception, IncomingMessage], Awaitable[None]]:
    async def exception_handler(
        message: IncomingMessage, bot: Bot, exc: Exception
    ) -> None:
        if not get_debug_enabled():
            log_incoming_message(
                message.raw_command,
                "Error while processing incoming message:",
                log_levels.ERROR,
            )
            flush_accumulated_logs(log_levels.ERROR)

        logger.exception(attach_log_source(""))

        await bot.answer_message(error_text)

    return exception_handler


async def fastapi_exception_handler(request: Request, exc: Exception) -> Response:
    if not get_debug_enabled():
        flush_accumulated_logs(log_levels.ERROR)
        log_incoming_http_request(
            "Error while processing incoming http request",
            log_levels.ERROR,
        )

    return Response(status_code=HTTP_500_INTERNAL_SERVER_ERROR)


async def smartapp_exception_handler(
    exc: Exception,
    smartapp: SmartApp,
) -> RPCErrorResponse:
    if not get_debug_enabled():
        log_system_event(
            smartapp.event.raw_command,
            "Error while processing incoming SmartApp event:",
            log_levels.ERROR,
        )
        flush_accumulated_logs(log_levels.ERROR)

    logger.exception(attach_log_source(""))

    return RPCErrorResponse(
        errors=[RPCError(reason="Internal error", id=exc.__class__.__name__.upper())],
    )



"""Singleton for `undefined` value."""


class Undefined:
    def __new__(cls) -> "Undefined":
        if not hasattr(cls, "instance"):  # noqa: WPS421
            cls.instance = super().__new__(cls)

        return cls.instance


undefined = Undefined()



"""Module for wrapping asyncio tasks with smart logger."""

from functools import wraps
from typing import Callable

from loguru import logger

from pybotx_smart_logger import log_levels
from pybotx_smart_logger.contextvars import (
    clear_accumulated_logs,
    set_debug_enabled,
    set_log_source,
)
from pybotx_smart_logger.logger import flush_accumulated_logs
from pybotx_smart_logger.schemas import LogSource


def make_smart_logger_decorator(
    debug_enabled_for_task: Callable[[str], bool]
) -> Callable:
    def decorator(func):  # type: ignore
        @wraps(func)
        async def wrapper(*args, **kwargs):  # type: ignore
            func_name = func.__name__

            set_debug_enabled(debug_enabled_for_task(func_name))
            set_log_source(LogSource.from_asyncio_task(func_name))
            clear_accumulated_logs()

            try:
                return await func(*args, **kwargs)
            except Exception:
                logger.error(f"An error happened in task '{func_name}':")
                flush_accumulated_logs(log_levels.ERROR)
                logger.exception("")

        return wrapper

    return decorator



from contextlib import contextmanager
from typing import Iterator

from pybotx.models.commands import SystemEvent
from loguru import logger

from pybotx_smart_logger import log_levels
from pybotx_smart_logger.contextvars import set_debug_enabled, set_log_source
from pybotx_smart_logger.logger import flush_accumulated_logs
from pybotx_smart_logger.output import log_system_event, attach_log_source
from pybotx_smart_logger.schemas import LogSource


@contextmanager
def wrap_system_event(
    event: SystemEvent,
    debug: bool = False,
) -> Iterator[None]:
    set_debug_enabled(debug)
    set_log_source(LogSource.from_system_event(event))

    if debug:
        log_system_event(
            event.raw_command,
            "Processing incoming event:",
            log_levels.INFO,
        )

    try:
        yield
    except Exception as exc:
        if not debug:
            log_system_event(
                event.raw_command,
                "Error while processing incoming SmartApp event:",
                log_levels.ERROR,
            )
            flush_accumulated_logs(log_levels.ERROR)
        
        logger.exception(attach_log_source(""))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pybotx_smart_logger']

package_data = \
{'': ['*']}

install_requires = \
['fastapi>=0.70.1,<0.71.0', 'pybotx>=0.32.0,<0.33.0']

setup_kwargs = {
    'name': 'pybotx-smart-logger',
    'version': '0.6.2',
    'description': 'Shows logs when you need it',
    'long_description': '# pybotx-smart-logger\n\n_Shows logs when you need it_\n\n\n## Проблема/решение\n\nВ основном наши боты работают в закрытых контурах. Там невозможно использовать Sentry,\nпоэтому наш главный помощник в диагностике неполадок -- логи контейнера.\n\nОднако, если сделать логи слишком подробными, то действительно важную информацию будет\nочень сложно найти. Также мы получим проблемы избыточного использования постоянной\nпамяти или слишком быструю ротацию логов. Но сделав логи слишком сжатыми, мы рискуем\nстолкнуться с ситуацией, когда их недостаточно для диагностики ошибки.\n\nТо есть хочется видеть как можно больше информации во время возникновения ошибок, и как\nможно меньше -- когда всё хорошо.\n\n\n## Использование\n\nИспользуя функцию `smart_log(log_message: str, log_item: Any = undefined)` логируете всю\nинформацию, которая поможет в диагностике ошибки. Если во время обработки сообщения\nбудет выброшено исключение, в логи попадёт:\n\n1. Текущее сообщение от пользователя,\n2. Вся залогированная с помощью `smart_log` информация,\n3. Выброшенное исключение.\n\nЕсли обработка сообщения завершится успешно, накопленные логи будут "выброшены".\n\nВ ботах часто используются фоновые задачи, исключения которых не могут быть перехвачены\nхендлером. Для них используется `smart_logger_decorator`, позволяющий получить\nаналогичное для обработки сообщений поведение.\n\nВы так же можете подлючить миддлварь для возможности логирования из FastAPI хендлеров. В таком случае, если во время обработки исключения будет выброшено ислючение, в логи попадет:\n\n1. Метод\n2. Урл запроса с query параметрами\n3. Заголовки запроса\n\nПри необходимости, тело запроса нужно логировать в самом хендлере:\n``` python\n`smart_log(await request.json())`\n```\n\n\n## Режим отладки\n\nПоявление необходимых логов при возникновении ошибки очень удобно. Однако есть\nситуации, когда необходимо понаблюдать за нормальной работой кода. Тем более, что вызовы\n`smart_log` уже расставлены. Поэтому для сообщений и фоновых задач предусмотрены функции\nотладки.\n\n`BotXSmartLoggerMiddleware` принимает пользовательскую функцию\n`debug_enabled_for_message(message: Message) -> bool`, позволяющую включить режим отладки в\nзависимости от входящего сообщения.\n\n`make_smart_logger_decorator` принимает пользовательскую функцию\n`debug_enabled_for_task(task_name: str) -> bool` позволяющую включить режим отладки в\nзависимости от имени функции. `smart_logger_decorator` знает имя функции, которую он\nоборачивает и передаёт его в `debug_enabled_for_task` в качестве аргумента.\n\nЭти функции можно использовать для включения отладки через переменные окружения, стейт\nбота, Redis, PostgreSQL и т.д. Рекомендую завести команды, которые позволят включать\nрежим отладки отправкой сообщения (см. пример ниже).\n\n`FastApiSmartLoggerMiddleware` принимает аргумент `debug_enabled: bool`.\n\n\n## Настройка\n\n1. Добавляем библиотеку в зависимости `poetry` и устанавливаем:\n\n```toml\npybotx-smart-logger = { git = "https://gitlab.ccsteam.ru/rnd/pybotx-smart-logger", rev = "0.6.0" }\n```\n\n```bash\n$ poetry lock && poetry install\n```\n\n2. Создаём хендлер исключений и подлючаем его к боту. Хендлер должен быть подключен к\n   типу Exception, т.е. заменяет подключенный в коробке `internal_error_handler`.\n\n```python\nfrom app.resources import strings\nfrom pybotx_smart_logger import make_smart_logger_exception_handler\n\nsmart_logger_exception_handler = make_smart_logger_exception_handler(\n    strings.BOT_INTERNAL_ERROR_TEXT\n)\n\nbot.add_exception_handler(Exception, smart_logger_exception_handler)\n```\n\n3. Подключаем миддлварь:\n\n```python\nfrom pybotx_smart_logger import BotXSmartLoggerMiddleware\n\nbot.add_middleware(\n    BotXSmartLoggerMiddleware, debug_enabled_for_message=lambda message: False\n)\n```\n\n4. [Опционально] Для фоновых задач создаём декоратор и запускаем фоновые задачи в при\n   старте бота:\n\n```python\nimport asyncio\nfrom pybotx_smart_logger import make_smart_logger_decorator\n\nsmart_logger_decorator = make_smart_logger_decorator(lambda task_name: False)\n\n\n@smart_logger_decorator\nasync def update_users_tasks() -> None:\n    pass\n\n\nasync def update_background_task() -> None:\n    while True:\n        await update_users_tasks()\n        await asyncio.sleep(60)\n\n# Внутри функции бота `start_app`:\nasyncio.create_task(update_background_task())\n```\n\n5. [Опционально] Возможность логирования из FastAPI хендлера:\nВ файле `app/main.py`\n5.1 Подлключаем миддлварь:\n``` python\nfrom pybotx_smart_logger import FastApiSmartLoggerMiddleware\n...\ndef get_application() -> FastAPI:\n    ...\n    application.middleware("http")(FastApiSmartLoggerMiddleware(debug_enabled=False))\n```\n5.2 Подключаем хендлер исключений:\n``` python\nfrom pybotx_smart_logger import FastApiSmartLoggerMiddleware, fastapi_exception_handler\n...\ndef get_application() -> FastAPI:\n    ...\n    application.middleware("http")(FastApiSmartLoggerMiddleware(debug_enabled=False))\n    application.add_exception_handler(Exception, fastapi_exception_handler)\n```\n\n## Пример команд для включения отладки\n\n```python\n@collector.hidden(command="/_debug:enable-for-huids")\nasync def enable_debug_for_users(message: Message, bot: Bot) -> None:\n    try:\n        huids = [UUID(huid) for huid in message.command.arguments]\n    except ValueError:\n        await bot.answer_message("Получен невалидный user_huid", message)\n        return\n\n    # TODO: Обновите список user_huid\n\n    await bot.answer_message("Список user_huid для отладки обновлён", message)\n```\n\n\n```python\n@collector.hidden(command="/_debug:enable-for-tasks")\nasync def enable_debug_for_tasks(message: Message, bot: Bot) -> None:\n    # TODO: Обновите список имён задач\n\n    await bot.answer_message("Список задач для отладки обновлён", message)\n```\n\n\n## Где применять\n\n1. Проверка роли:\n\n```python\nfrom pybotx_smart_logger import smart_log\n\nasync def subscribed_users_only(\n    message: Message,\n    bot: Bot,\n    user: User = user_dependency,\n) -> None:\n    if not user.is_subscribed:\n        await bot.send(only_subscribed_users_allowed_message(message))\n\n        raise DependencyFailure\n\n    smart_log("This user is subscribed")\n```\n\n2. Обращение в API:\n\n```python\nfrom pybotx_smart_logger import smart_log\n\nasync def _perform_request(\n    self,\n    method: Literal["GET", "POST"],\n    url: str,\n    query_params: Optional[Dict[str, Any]] = None,\n    body_dict: Optional[Dict[str, Any]] = None,\n) -> ResponseSchema:\n    smart_log("Performing request to YourAwesomeAPI")\n    smart_log("Method:", method)\n    smart_log("URL:", url)\n    smart_log("Query parameters:", query_params)\n    smart_log("Body dict:", body_dict)\n\n    try:\n        async with AsyncClient(base_url=self._base_url) as client:\n            response = await client.request(\n                method, url, params=query_params, json=body_dict\n            )\n    except HTTPError as exc:\n        raise RequestToAwesomeAPIFailed from exc\n\n    smart_log("Response text:", response.text)\n\n    try:\n        response.raise_for_status()\n    except HTTPStatusError as exc:  # noqa: WPS440\n        raise InvalidStatusCodeFromAwesomeAPI from exc\n\n    return response.text\n```\n\nА также любые моменты, где что-то может пойти не так. Логируйте -- не стестяйтесь.\n',
    'author': 'Alexander Samoylenko',
    'author_email': 'alexandr.samojlenko@ccsteam.ru',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



"""Prints logs only if exception was raised while processing message."""

import inspect
from pprint import pformat
from typing import Any

from loguru import logger

from pybotx_smart_logger import log_levels
from pybotx_smart_logger.contextvars import (
    clear_accumulated_logs,
    get_accumulated_logs,
    get_debug_enabled,
)
from pybotx_smart_logger.output import attach_log_source
from pybotx_smart_logger.schemas import LogEntry
from pybotx_smart_logger.undefined import Undefined, undefined


def flush_log_entry(log_entry: LogEntry, log_level: str) -> None:
    patched_logger = logger.patch(
        lambda record: record.update(
            {
                "name": log_entry.module,
                "line": log_entry.line_number,
                "function": log_entry.function,
            }
        )
    )

    formatted_log_message = attach_log_source(log_entry.log_message)

    if isinstance(log_entry.log_item, Undefined):
        patched_logger.log(log_level, formatted_log_message)
    else:
        patched_logger.log(
            log_level, "{}\n{}", formatted_log_message, pformat(log_entry.log_item)
        )


def flush_accumulated_logs(log_level: str) -> None:
    for log_entry in get_accumulated_logs():
        flush_log_entry(log_entry, log_level)

    clear_accumulated_logs()


def smart_log(log_message: str, log_item: Any = undefined) -> None:
    caller_frame = inspect.stack()[1]
    log_entry = LogEntry(
        log_message=log_message,
        log_item=log_item,
        module=caller_frame.frame.f_globals["__name__"],
        function=caller_frame.function,
        line_number=caller_frame.lineno,
    )

    if get_debug_enabled():
        flush_log_entry(log_entry, log_levels.INFO)
    else:
        accumulated_logs = get_accumulated_logs()
        accumulated_logs.append(log_entry)
